import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


public class Chainer extends EffectFilter
{
	class MassPoint
	{
		public MassPoint()
		{
			x = 0;
			y = 0;
			vx = 0;
			vy = 0;
			mass = 1.0;
			fixed = false;
		}
		
		public MassPoint(MassPoint src)
		{
			x = src.x;
			y = src.y;
			vx = src.vx;
			vy = src.vy;
			mass = src.mass;
			fixed = src.fixed;
		}
		
		double x;
		double y;
		double vx;
		double vy;
		double mass;
		boolean fixed;
	}
	
	class MassChain
	{
		private static final double GRAVITY_X = 0.0;
		private static final double GRAVITY_Y = 5.0;
		
		private static final double ATTRACTION_FACTOR = 10.0;
		private static final double ATTRACTION_DISTANCE = 100.0;
		
		private static final double DAMPING = 0.3;
		
		ArrayList<MassPoint> nodes;
		
		// how long the links naturally are
		double nLength;
		
		// how far the links can stretch before breaking
		double strength;
		int maxNodes;
		
		boolean attached;
		boolean broken;
		boolean dead;
		
		// location that it is 'attached' to (not fixed end)
		double ax;
		double ay;
		
		// spring coefficient
		double k;
		
		public MassChain(double seedX, double seedY, double nstrength, int mnodes, double length, double nk)
		{
			strength = nstrength;
			maxNodes = mnodes;
			
			attached = false;
			broken = false;
			dead = false;
			
			nLength = length;
			
			ax = -1.0;
			ay = -1.0;
			
			k = 1.0 * nk;
			
			nodes = new ArrayList<MassPoint>();
			
			MassPoint firstNode = new MassPoint();
			
			firstNode.x = seedX;
			firstNode.y = seedY;
			firstNode.vx = 0;
			firstNode.vy = 0;
			firstNode.mass = 1.0;
			firstNode.fixed = true;
			
			nodes.add(firstNode);
		}
		
		public void attractToContours(ArrayList<OUtil.Contour> contours, double dt)
		{
			if(broken)
				return;
			
			OUtil.Contour cur;
			double minDist = 10000.0;
			OUtil.Point closestPoint = null;
			
			double dist, dx, dy, fx, fy, force;
			OUtil.Point curPoint;
			
			MassPoint last = nodes.get(nodes.size() - 1);
			
			for(int i = 0; i < contours.size(); ++i)
			{
				cur = contours.get(i);
				
				if(cur == null)
					continue;
					
				for(int j = 0; j < cur.path.size(); ++j)
				{
					curPoint = cur.path.get(j);
					
					dx = curPoint.x - last.x;
					dy = curPoint.y - last.y;
					dist = Math.sqrt((dx * dx) + (dy * dy));
					
					if(dist < minDist)
					{
						minDist = dist;
						closestPoint = curPoint;
					}
				}
			}
			
			if(minDist > 5.0)
			{
				if(minDist <= ATTRACTION_DISTANCE)
					force = ATTRACTION_FACTOR;
				else
					force = 0.0;
			}
			else
			{
				//System.out.println("Attached!");
				attached = true;
				force = 0.0;
			}
			

			if(closestPoint == null)
				return;
			
			if(!attached)
			{
				double tx = closestPoint.x - last.x;
				double ty = closestPoint.y - last.y;
		
				double ux = tx / minDist;
				double uy = ty / minDist;
				
				fx = force * ux;
				fy = force * uy;
				
				last.vx += (fx / last.mass) * dt;
				last.vy += (fy / last.mass) * dt;
				
				// We shouldn't be damping here
				//last.vx -= DAMPING * last.vx * dt;
				//last.vy -= DAMPING * last.vy * dt;
			}
			else // attached
			{
				last.fixed = true;
				last.x = closestPoint.x;
				last.y = closestPoint.y;
			}
		}
		
		public void drawChain(FloatImage dest)
		{
			MassPoint cur;
			MassPoint prev;
			
			int skip;
			
			if(broken)
				skip = 2;
			else
				skip = 1;
			
			if(nodes.size() < 2)
				return;
			
			for(int i = 1; i < nodes.size(); i += skip)
			{
				prev = nodes.get(i-1);
				cur = nodes.get(i);
				
				OUtil.drawLine(dest, (int)cur.x, (int)prev.x, (int)cur.y, (int)prev.y, 1.0);
			}
		}
		
		public void grow()
		{
			if(nodes.size() >= maxNodes)
				return;
			
			MassPoint addee = new MassPoint();
			MassPoint last;
			
			if(nodes.size() == 0)
			{
				addee.x = 50;
				addee.y = 10;
				addee.vx = 0;
				addee.vy = 0;
				addee.mass = 1.0;
				addee.fixed = true;
			}
			else
			{
				last = nodes.get(nodes.size() - 1);
				
				addee.x = last.x + (Math.random() * 40.0) - 20.0;
				addee.y = last.y + nLength;
				addee.vx = 0;
				addee.vy = 0;
				addee.mass = 1.0;
				addee.fixed = false;
			}
			
			nodes.add(addee);
		}
		
		public boolean isDead()
		{
			return dead;
		}
		
		public void update(double dt)
		{
			int skip;
			double maxy = 0.0;
			
			if(broken)
				skip = 2;
			else
				skip = 1;
			
			// first pass: compute new velocities
			
			// force in x and y axes
			double fx, fy;
			
			// acceleration in x and y aces
			double acx, acy;
			
			// vector from cur to next
			double tx, ty;
			
			// unit vector from cur to next
			double ux, uy;
			
			// magnitude
			double dist;
			
			MassPoint cur, next;
			
			for(int i = 0; i < nodes.size() - 1; i += skip)
			{
				cur = nodes.get(i);
				next = nodes.get(i+1);
				
				tx = next.x - cur.x;
				ty = next.y - cur.y;
				
				dist = Math.sqrt(tx*tx + ty*ty);
				
				if(dist <= 0)
					continue;
					
				if(dist > strength && !broken)
				{
					System.out.println("Chain broke!");
					broken = true;
					
					ArrayList<MassPoint> brokenChain = new ArrayList<MassPoint>();
					
					nodes.get(0).fixed = false;
					brokenChain.add(nodes.get(0));
					
					MassPoint curPnt;
					
					for(int j = 1; j < nodes.size()-1; ++j)
					{
						curPnt = nodes.get(j);
							
						curPnt.fixed = false;
						curPnt.vx += (Math.random() * 40.0) - 20.0;
						curPnt.vy += (Math.random() * 40.0) - 20.0;
						brokenChain.add(curPnt);
						brokenChain.add(new MassPoint(curPnt));
					}
					
					nodes.get(nodes.size()-1).fixed = false;
					brokenChain.add(nodes.get(nodes.size()-1));
					
					nodes = brokenChain;
					
					break;
				}
		
				ux = tx / dist;
				uy = ty / dist;
				
				fx = k * (dist - nLength) * ux;
				fy = k * (dist - nLength) * uy;
				
				if(!cur.fixed)
				{
					acx = fx / cur.mass;
					acy = fy / cur.mass;
					
					// we only add gravity here so that it isn't double added
					cur.vx += (acx + GRAVITY_X) * dt;
					cur.vy += (acy + GRAVITY_Y) * dt;
				}
				
				if(!next.fixed)
				{
					acx = -1.0 * fx / next.mass;
					acy = -1.0 * fy / next.mass;
					
					next.vx += (acx * dt);
					next.vy += (acy * dt);
					
					// because of the skip, if it's broken, we need to add gravity
					//  here as well
					
					if(broken)
					{
						next.vx += (acx + GRAVITY_X) * dt;
						next.vy += (acy + GRAVITY_Y) * dt;
					}
				}
					
			}
			
			// second pass: compute new positions
			for(int i = 0; i < nodes.size(); ++i)
			{
				cur = nodes.get(i);
				
				// we only damp here to make sure we don't double damp
				cur.vx -= DAMPING * cur.vx * dt;
				cur.vy -= DAMPING * cur.vy * dt;
				
				cur.x += (cur.vx * dt);
				cur.y += (cur.vy * dt);
				
				if(cur.y > maxy)
					maxy = cur.y;
			}
			
			if(maxy >= 500.0)
				dead = true;
		}
	}
	
	protected transient ArrayList<MassChain> chains;
	protected transient OUtil.EdgeField efield;
	
	protected static final int NUM_CHAINS = 20;
		
	public Chainer()
	{
		filterInputs = new FilterLink[1];
		filterInputs[0] = null;
		dead = false;
		filterName = "Chainer " + (int)(Math.random() * 1000);
		
		guiWidth = 100;
		guiHeight = 100;
		guiX = 200;
		guiY = 200;
		
		rebuildFromSave();
	}
	
	public void rebuildFromSave()
	{
		internalBuffer = new FloatImage();
		efield = new OUtil.EdgeField();
		
		remakeChains(320);
		
		needsProcessing = true;
		
		guiWindow = new JInternalFrame(filterName, true, true, true, true);
		guiWindow.setSize(guiWidth,guiHeight);
		guiWindow.setLocation(guiX, guiY);
		
		guiWindow.add(new JLabel("Chainer"));
		
		guiWindow.setVisible(true);
	}
	
	public void remakeChains(int width)
	{	
		chains = new ArrayList<MassChain>();
		
		for(int i = 0; i < NUM_CHAINS; ++i)
		{
			chains.add(makeNewChain(width));
		}
	}
	
	public MassChain makeNewChain(int width)
	{
		int chainX;
		int chainLength;
		
		chainX = (int)(Math.random() * width);
		chainLength = (int)(Math.random() * 10) + 3;
		
		//                    seedX  Y   str     #nodes    length  k
		return new MassChain(chainX, 10, 40.0, chainLength, 3.0, 1.0);	
	}
	
	public FloatImage getOutput(int outputChannel)
	{	
		if(needsProcessing && filterInputs[0] != null)
		{
			FloatImage temp = filterInputs[0].getOutput();
			
			if(temp != null)
			{
				if(temp.getWidth() != internalBuffer.getWidth() || temp.getHeight() != internalBuffer.getHeight())
				{
					internalBuffer = new FloatImage(3, temp.getWidth(), temp.getHeight());
					remakeChains(temp.getWidth());
				}
				
				int imgWidth = internalBuffer.getWidth();
				int imgHeight = internalBuffer.getHeight();
				
				double[][] ored = internalBuffer.getChannelData(0);
				double[][] ogreen = internalBuffer.getChannelData(1);
				double[][] oblue = internalBuffer.getChannelData(2);
				
				double min;
				
				// first, fill with black
				for(int i = 0; i < imgWidth; ++i)
				{
					for(int j = 0; j < imgHeight; ++j)
					{
						ored[i][j] = -1.0;
						ogreen[i][j] = -1.0;
						oblue[i][j] = -1.0;
					}
				}
				
				efield.makeEdgeField(temp);
				
				ArrayList<OUtil.Contour> outlines = efield.getOutlines();
				ArrayList<OUtil.Contour> smoothedOutlines = new ArrayList<OUtil.Contour>();
				
				OUtil.Contour tempC;
				
				OUtil.Point curPoint;
				
				int ix, iy, cc, maxc;
				
				double dc;
				double dcr, dcg;
				
				if(outlines == null)
				{
					System.out.println("Null outline returned from get outlines");
					return internalBuffer;
				}
				
				if(outlines.size() == 0)
					System.out.println("No outlines!");
					
				//System.out.println("Num outlines: " + outlines.size());
				
				maxc = 0;
				
				for(int i = 0; i < outlines.size(); ++i)
				{
					tempC = outlines.get(i);
					
					cc = 0;
					
					tempC = OUtil.downsampleContour(tempC, 5);
					
					tempC = OUtil.smoothContour(tempC, 10);
					tempC = OUtil.smoothContour(tempC, 10);
					tempC = OUtil.smoothContour(tempC, 10);
					
					if(tempC == null || tempC.path == null)
						continue;
						
					smoothedOutlines.add(tempC);
					
					for(int j = 0; j < tempC.path.size(); ++j)
					{
						curPoint = tempC.path.get(j);
						
						ix = (int)curPoint.x;
						iy = (int)curPoint.y;
						
						if(ix < 0)
							ix = 0;
						if(ix >= imgWidth)
							ix = imgWidth-1;
						if(iy < 0)
							iy = 0;
						if(iy >= imgHeight)
							iy = imgHeight-1;
						
						ored[ix][iy] = 1.0;
						ogreen[ix][iy] = 1.0;
						oblue[ix][iy] = 1.0;
					}
				}
				
				MassChain tempChain;
				
				for(int i = 0; i < chains.size(); ++i)
				{
					tempChain = chains.get(i);
					tempChain.grow();
					tempChain.update(1.0);
					tempChain.attractToContours(smoothedOutlines, 1.0);
					tempChain.drawChain(internalBuffer);
					
					if(tempChain.isDead())
					{
						chains.set(i, makeNewChain(imgWidth));
						System.out.println("Generated new chain");
					}
				}
				
				//System.out.println("Max path length: " + maxc);
			}
			needsProcessing = false;
		}
		
		return internalBuffer;
	}	
}