import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import javax.swing.*;

public class CombinedCalibrationFilter extends EffectFilter implements ChangeListener
{		
	// sliders for translation and scaling
	private transient JSlider xTranslateSlider;
	private transient JSlider yTranslateSlider;
	private transient JSlider xScaleSlider;
	private transient JSlider yScaleSlider;
	
	// slider for rotation
	private transient JSlider thetaSlider;
	
	// fields for lens correction
	protected transient JFormattedTextField aField;
	protected transient JFormattedTextField bField;
	protected transient JFormattedTextField cField;
	protected transient JFormattedTextField dField;
	
	protected IntPoint[][] transformTable;
	
	// variables for translation and scaling
	private double translateX;
	private double translateY;
	private double scaleX;
	private double scaleY;
	
	// variable for rotation
	private double rotationAmount;
	
	// variables for lens distortion parameters
	private double ap;
	private double bp;
	private double cp;
	private double dp;
	
	// controls how precise the sliders are
	private static final double GRANULARITY = 1000.0;
	
	private boolean rebuildTable;

	public CombinedCalibrationFilter()
	{
		filterInputs = new FilterLink[1];
		filterInputs[0] = null;
		dead = false;
		filterName = "Manual Calibration Filter " + (int)(Math.random() * 1000);
		
		guiWidth = 100;
		guiHeight = 100;
		guiX = 150;
		guiY = 150;
		
		// initial translation
		translateX = 0.0;
		translateY = 0.0;
		scaleX = 1.0;
		scaleY = 1.0;
		
		// initial rotation
		rotationAmount = 0.0;
		
		// initial lens parameters
		ap = 0.0;
		bp = 0.0;
		cp = 0.0;
		dp = 1.0;
		
		rebuildTable = true;
		
		rebuildFromSave();
	}
	
	public void prepareForSave()
	{	
		super.prepareForSave();
		getSliderValues();
	}
	
	public void getSliderValues()
	{
		// get translation/scaling
		translateX = xTranslateSlider.getValue();
		translateY = yTranslateSlider.getValue();
		
		scaleX = (double)xScaleSlider.getValue() / 100.0;
		scaleY = (double)yScaleSlider.getValue() / 100.0;
		
		// get rotation
		rotationAmount = (double)thetaSlider.getValue() / GRANULARITY;
		
		// get lens parameters
		ap = (Double)aField.getValue();
		bp = (Double)bField.getValue();
		cp = (Double)cField.getValue();
		dp = (Double)dField.getValue();
	}
	
	public void setSliderValues()
	{
		System.out.println("Set slider values");
		
		// set translation/scaling
		xTranslateSlider.setValue((int)translateX);
		yTranslateSlider.setValue((int)translateY);
		
		xScaleSlider.setValue((int)(scaleX * 100.0));
		yScaleSlider.setValue((int)(scaleY * 100.0));
		
		// set rotation
		thetaSlider.setValue((int)(rotationAmount * GRANULARITY));
		
		// there is no set for the lens parameters: they should all be saved
		//  elsewhere (hopefully)
	}
	
	public void stateChanged(ChangeEvent e)
	{
		// indicate that the table should be rebuilt
		rebuildTable = true;
	}
	
	public void rebuildFromSave()
	{
		internalBuffer = new FloatImage();
		needsProcessing = true;
		rebuildTable = true;
		
		guiWindow = new JInternalFrame(filterName, true, true, true, true);
		guiWindow.setSize(guiWidth, guiHeight);
		guiWindow.setLocation(guiX, guiY);
		
		// tabs for the individual calibration things
		JTabbedPane tabs = new JTabbedPane();
		
		JPanel whatever = new JPanel();
		
		xTranslateSlider = new JSlider(JSlider.HORIZONTAL);
		xTranslateSlider.setMajorTickSpacing(20);
		xTranslateSlider.setMinorTickSpacing(10);
		xTranslateSlider.setPaintTicks(true);
		xTranslateSlider.setBorder(BorderFactory.createTitledBorder("X Translation"));
		xTranslateSlider.setMinimum(-200);
		xTranslateSlider.setMaximum(200);
		xTranslateSlider.addChangeListener(this);
		
		whatever.add(xTranslateSlider);
		
		yTranslateSlider = new JSlider(JSlider.HORIZONTAL);
		yTranslateSlider.setMajorTickSpacing(20);
		yTranslateSlider.setMinorTickSpacing(10);
		yTranslateSlider.setPaintTicks(true);
		yTranslateSlider.setBorder(BorderFactory.createTitledBorder("Y Translation"));
		yTranslateSlider.setMinimum(-200);
		yTranslateSlider.setMaximum(200);
		yTranslateSlider.addChangeListener(this);
		
		whatever.add(yTranslateSlider);
		
		xScaleSlider = new JSlider(JSlider.HORIZONTAL);
		xScaleSlider.setMajorTickSpacing(20);
		xScaleSlider.setMinorTickSpacing(10);
		xScaleSlider.setPaintTicks(true);
		xScaleSlider.setBorder(BorderFactory.createTitledBorder("X Scaling"));
		xScaleSlider.setMinimum(-200);
		xScaleSlider.setMaximum(200);
		xScaleSlider.addChangeListener(this);
		
		whatever.add(xScaleSlider);
		
		yScaleSlider = new JSlider(JSlider.HORIZONTAL);
		yScaleSlider.setMajorTickSpacing(20);
		yScaleSlider.setMinorTickSpacing(10);
		yScaleSlider.setPaintTicks(true);
		yScaleSlider.setBorder(BorderFactory.createTitledBorder("Y Scaling"));
		yScaleSlider.setMinimum(-200);
		yScaleSlider.setMaximum(200);
		yScaleSlider.addChangeListener(this);
		
		whatever.add(yScaleSlider);
		
		tabs.addTab("Translation", whatever);
		
		// now we make the tab for rotation
		whatever = new JPanel();
		
		thetaSlider = new JSlider(JSlider.HORIZONTAL);
		thetaSlider.setMajorTickSpacing(200);
		thetaSlider.setMinorTickSpacing(100);
		thetaSlider.setPaintTicks(true);
		thetaSlider.setBorder(BorderFactory.createTitledBorder("Rotation"));
		thetaSlider.setMinimum(-2 * (int)GRANULARITY);
		thetaSlider.setMaximum(2 * (int)GRANULARITY);
		thetaSlider.addChangeListener(this);
		
		whatever.add(thetaSlider);
		
		tabs.addTab("Rotation", whatever);
		
		// finally, we deal with lens correction
		JPanel foo = new JPanel();
		
		aField = new JFormattedTextField(new Double(ap));
		aField.setBorder(BorderFactory.createTitledBorder("a"));
		aField.setPreferredSize(new Dimension(200, 50));
		
		bField = new JFormattedTextField(new Double(bp));
		bField.setBorder(BorderFactory.createTitledBorder("b"));
		bField.setPreferredSize(new Dimension(200, 50));
		
		cField = new JFormattedTextField(new Double(cp));
		cField.setBorder(BorderFactory.createTitledBorder("c"));
		cField.setPreferredSize(new Dimension(200, 50));
		
		dField = new JFormattedTextField(new Double(dp));
		dField.setBorder(BorderFactory.createTitledBorder("d"));
		dField.setPreferredSize(new Dimension(200, 50));
		
		foo.add(aField);
		foo.add(bField);
		foo.add(cField);
		foo.add(dField);
		
		JButton calculateButton = new JButton("Make Transform");
		calculateButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e)
			{
				getSliderValues();
				buildTable();
			}
		});
		foo.add(calculateButton);
		
		tabs.addTab("Lens Distortion", foo);
		
		setSliderValues();
		
		buildTable();
		
		guiWindow.add(tabs);
		
		guiWindow.setVisible(true);
	}
	
	public String getName()
	{
		return filterName;
	}
	
	public String getInputName(int inputNum)
	{
		return "Input";
	}
	
	public String getOutputName(int outputNum)
	{
		return "Output";
	}
	
	public int numOutputs()
	{
		return 1;
	}
	
	public void buildTable()
	{
		// here is the order: input -> lens correction -> rotation -> T/S
		// therefore, to reuse the old code we have to go backwards: starting
		// with the final destination pixel we run the T/S transform, then the
		// rotation transform, and finally the lens correction transform to get
		// the source pixel in the input image
		
		System.out.println("TX: " + translateX);
		System.out.println("TY: " + translateY);
		System.out.println("SX: " + scaleX);
		System.out.println("SY: " + scaleY);	
	
		int imgWidth = internalBuffer.getWidth();
		int imgHeight = internalBuffer.getHeight();
		
		transformTable = new IntPoint[imgWidth][imgHeight];
		
		double cx = (double)imgWidth / 2.0;
		double cy = (double)imgHeight / 2.0;
		
		// rotation amound is in PI radians
		double theta = Math.PI * rotationAmount;
		
		// normalized radius: radiusn = 1.0
		double radiusn = imgHeight / 2.0;
		
		int ti, tj;
		double dx, dy, tdx, tdy, ii, ij, ii2, ij2;
		double dr, dr2, dr3, dr4, sr;
		double scaleFactor;
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				
				dx = (double)i;
				dy = (double)j;
				
				tdx = (dx * scaleX);
				tdy = (dy * scaleY);
				
				// intermediate coordinates after lens and rotation, but
				//  before T/S
				ii = translateX + tdx;
				ij = translateY + tdy;
				
				// now we transform for rotation
				dx = ii - cx;
				dy = ij - cy;
				
				tdx = (dx * Math.cos(theta) - dy * Math.sin(theta));
				tdy = (dx * Math.sin(theta) + dy * Math.cos(theta));
				
				// intermediate coordinates after lens but before anything else
				ii2 = (cx + tdx);
				ij2 = (cy + tdy);
				
				// now we transform for lens correction
				dx = ii2 - cx;
				dy = ij2 - cy;
				
				dr = Math.sqrt((dx * dx) + (dy * dy)) / radiusn;
				dr2 = dr * dr;
				dr3 = dr * dr2;
				dr4 = dr2 * dr2;
				
				sr = (ap * dr4) + (bp * dr3) + (cp * dr2) + (dp * dr);
				
				if(dr > 0)
					scaleFactor = sr / dr;
				else
					scaleFactor = 0;
				
				ti = (int)(cx + (dx * scaleFactor));
				tj = (int)(cy + (dy * scaleFactor));
				
				if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
				{
					transformTable[i][j] = new IntPoint(ti, tj);
				}
			}
		}
		
		System.out.println("Table built");
	}
	
	public FloatImage getOutput(int outputChannel)
	{
		if(rebuildTable == true)
		{
			getSliderValues();
			buildTable();
			rebuildTable = false;
			needsProcessing = true;
		}
		
		if(needsProcessing && filterInputs[0] != null)
		{
			FloatImage temp = filterInputs[0].getOutput();
			
			if(temp != null)
			{	
				if(temp.getWidth() != internalBuffer.getWidth() || temp.getHeight() != internalBuffer.getHeight())
				{
					internalBuffer = new FloatImage(3, temp.getWidth(), temp.getHeight());
					
					// if the input size changes, it invalidates the transform table
					transformTable = null;
				}
				
				int imgWidth = internalBuffer.getWidth();
				int imgHeight = internalBuffer.getHeight();
				
				double[][] ored = internalBuffer.getChannelData(0);
				double[][] ogreen = internalBuffer.getChannelData(1);
				double[][] oblue = internalBuffer.getChannelData(2);
				
				double[][] ired = temp.getChannelData(0);
				double[][] igreen = temp.getChannelData(1);
				double[][] iblue = temp.getChannelData(2);
				
				if(transformTable != null)
				{
					int ti, tj;
					
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							if(transformTable[i][j] != null)
							{
								ti = transformTable[i][j].x;
								tj = transformTable[i][j].y;
								
								if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
								{
									//System.out.println("T");
									
									ored[i][j] = ired[ti][tj];
									ogreen[i][j] = igreen[ti][tj];
									oblue[i][j] = iblue[ti][tj];
								}
								else
								{
									ored[i][j] = -1.0;
									ogreen[i][j] = -1.0;
									oblue[i][j] = -1.0;
								}
							}
							else
							{
								ored[i][j] = -1.0;
								ogreen[i][j] = -1.0;
								oblue[i][j] = -1.0;
							}
						}
					}
				}
				else // transformTable == null
				{
					System.out.println("Null tt");
					
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							ored[i][j] = ired[i][j];
							ogreen[i][j] = igreen[i][j];
							oblue[i][j] = iblue[i][j];
						}
					}
				}
			}
			needsProcessing = false;
		}

		return internalBuffer;
	}	
}