import jpaul.DataStructs.UnionFind;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

public class ConnectedComponentsFilter extends EffectFilter
{
	private Integer[][] intImage;
	private Integer[][] idImage;
	private int imgWidth;
	private int imgHeight;
	
	private transient HashMap<Integer, Integer> idToSize;
	private transient HashMap<Integer, Integer> idToRank;
	
	private static final int NUM_LEVELS = 4;
	
	class IdSizeEntry implements Comparable
	{
		public int id;
		public int size;
		
		public IdSizeEntry(int nid, int nsize)
		{
			id = nid;
			size = nsize;
		}
		
		public int compareTo(Object o)
		{
			if(!(o instanceof IdSizeEntry))
				return 0;
			
			IdSizeEntry ido = (IdSizeEntry)o;
			
			if(size > ido.size)
				return -1;
			else if(size == ido.size)
				return 0;
			else
				return 1;
		}
	}
	
	public ConnectedComponentsFilter()
	{
		filterInputs = new FilterLink[1];
		filterInputs[0] = null;
		dead = false;
		filterName = "Connected Components " + (int)(Math.random() * 1000);
		
		guiWidth = 100;
		guiHeight = 100;
		guiX = 200;
		guiY = 200;
		
		rebuildFromSave();
	}
	
	public void rebuildFromSave()
	{
		internalBuffer = new FloatImage();
		needsProcessing = true;
		
		guiWindow = new JInternalFrame(filterName, true, true, true, true);
		guiWindow.setSize(guiWidth,guiHeight);
		guiWindow.setLocation(guiX, guiY);
		
		guiWindow.add(new JLabel("Connected Components"));
		
		guiWindow.setVisible(true);
	}
	
	private void findComponents()
	{
		UnionFind<Integer> uf = new UnionFind<Integer>();
		
		Integer current, left, top;
		
		int currentId = 0;
		
		// first we find horizontal runs of like-colored pixels while building
		//  up the union find structure
		for(int j = 0; j < imgHeight; ++j)
		{
			for(int i = 0; i < imgWidth; ++i)
			{	
				idImage[i][j] = null;
			
				current = intImage[i][j];
				
				if(i > 0) // we compare the current pixel to the left one:
				{
					left = intImage[i-1][j];
					
					// if so, it gets left's id
					if(current.equals(left))
					{
						idImage[i][j] = idImage[i-1][j];
					}
					else // otherwise, it gets a new id
					{
						idImage[i][j] = new Integer(++currentId);
					}
				}
				else // no left pixel to compare it to, so it gets a new id by
				{	 // default
					idImage[i][j] = new Integer(++currentId);
				}
				
				if(j > 0) // now we compare it to the pixel above it:
				{
					top = intImage[i][j-1];	
					
					if(current.equals(top)) // if so, we union their ids
					{
						uf.union(idImage[i][j], idImage[i][j-1]);
					}
				}
			}
		}
		
		// now we run through a second pass in which every id becomes its
		//  representative, and we build up a hash table that indicates the
		//  size of each component
		
		idToSize = new HashMap<Integer, Integer>();
		
		Integer curSize;
		
		for(int j = 0; j < imgHeight; ++j)
		{
			for(int i = 0; i < imgWidth; ++i)
			{
				idImage[i][j] = uf.find(idImage[i][j]);
				
				curSize = idToSize.get(idImage[i][j]);
				
				if(curSize != null)
				{
					idToSize.put(idImage[i][j], curSize.intValue() + 1);
				}
				else
				{
					idToSize.put(idImage[i][j], 1);
				}
			}
		}
		
		// now we run through the hash, producing a list of ids and sizes that
		//  can then be sorted to get a ranking
		
		Set<Map.Entry<Integer, Integer>> entries = idToSize.entrySet();
		
		Map.Entry<Integer, Integer> cur = null;
		
		ArrayList<IdSizeEntry> ranking = new ArrayList<IdSizeEntry>();
		
		for(Iterator<Map.Entry<Integer, Integer>> iterator = entries.iterator(); iterator.hasNext();)
		{
			cur = iterator.next();
			
			ranking.add(new IdSizeEntry(cur.getKey(), cur.getValue()));
		}
		
		// now we sort the ranking...
		Collections.sort(ranking);
		
		// finally, we iterate through the ranking, re-adding all of the
		// things to the id to rank hash
		idToRank = new HashMap<Integer, Integer>();
		
		IdSizeEntry cur2 = null;
		int rank = 1;
		
		for(Iterator<IdSizeEntry> iterator = ranking.iterator(); iterator.hasNext();)
		{
			cur2 = iterator.next();
			
			idToRank.put(cur2.id, rank);
			
			++rank;
		}
		
	}
	
	public FloatImage getOutput(int outputChannel)
	{
		final double LEVELSD2 = (double)NUM_LEVELS / 2.0;
		
		if(needsProcessing && filterInputs[0] != null)
		{
			FloatImage temp = filterInputs[0].getOutput();
			
			if(temp != null)
			{
				imgWidth = internalBuffer.getWidth();
				imgHeight = internalBuffer.getHeight();
				
				if(temp.getWidth() != imgWidth || temp.getHeight() != imgHeight)
				{
					imgWidth = temp.getWidth();
					imgHeight = temp.getHeight();
					
					internalBuffer = new FloatImage(3, imgWidth, imgHeight);
					intImage = new Integer[imgWidth][imgHeight];
					idImage = new Integer[imgWidth][imgHeight];	
				}
				
				// we now create the intImage by just leveling the red channel
				//  of the source
				
				double[][] ired = temp.getChannelData(0);
				
				for(int j = 0; j < imgHeight; ++j)
				{
					for(int i = 0; i < imgWidth; ++i)
					{
						intImage[i][j] = new Integer((int)((ired[i][j] + 1.0) * LEVELSD2));
					}
				}
				
				findComponents();
				
				// finally, we translate from component ids into colors
				double[][] ored = internalBuffer.getChannelData(0);
				double[][] ogreen = internalBuffer.getChannelData(1);
				double[][] oblue = internalBuffer.getChannelData(2);
				
				// this is a gigantic hack to produce at least consistent colors
				//  out of the connected components info
				int r, g, b;
				
				for(int j = 0; j < imgHeight; ++j)
				{
					for(int i = 0; i < imgWidth; ++i)
					{
						r = idToRank.get(idImage[i][j]) % 100;
						
						ored[i][j] = ((double)r * 0.005) - 1.0;
						ogreen[i][j] = ((double)r * 0.005) - 1.0;
						oblue[i][j] = ((double)r * 0.005) - 1.0;
					}
				}
			}
			
			needsProcessing = false;
		}
		
		return internalBuffer;
	}	
}