public class ImagePlane
{
	protected FTriple p0;
	protected FTriple n;
	protected FTriple u;
	protected FTriple v;
	protected double aspect;
	protected double pixelWidth;
	protected double F;
	protected double pixelHeight;
	protected double pixelHalfWidth;
	protected double pixelHalfHeight;
	protected double halfWidth;
	protected double halfHeight;
	
	// it is assumed that the camera is at (0,0,0) facing in the -z-direction
	public ImagePlane(FTriple p0, FTriple n, FTriple u, FTriple v, double F, double aspect, double pixelHeight)
	{
		this.p0 = new FTriple(p0);
		this.n = n.norm();
		this.u = u.norm();
		this.v = v.norm();
		this.F = F;
		this.aspect = aspect;
		this.pixelHeight = pixelHeight;
		this.pixelWidth = pixelHeight * aspect;
		
		halfWidth = aspect / 2.0;
		halfHeight = 0.5; // since we define the total height as 1.0
		
		pixelHalfWidth = pixelWidth / 2.0;
		pixelHalfHeight = pixelHeight / 2.0;
		
	}
	
	// assumes x and y are in pixels, with (0,0) at the center of the screen,
	// +x to the right and +y up
	public FTriple XYtoUV(FTriple XY)
	{
		if(XY == null)
			return null;
		
		// first we convert pixel coordinates to image plane coordinates
		double ix = (XY.val[0] / pixelHalfWidth) * halfWidth;
		double iy = (XY.val[1] / pixelHalfHeight) * halfHeight;
		
		// now we need to calculate the 'sight line' through that point
		FTriple right 	= new FTriple( 1.0,  0.0,  0.0);
		FTriple up 		= new FTriple( 0.0,  1.0,  0.0);
		FTriple forward = new FTriple( 0.0,  0.0, -1.0); 
		
		// direction = (right * ix) + (up * iy) + (forward * F)
		FTriple direction = ((right.times(ix)).plus(up.times(iy))).plus(forward.times(F));
		
		// now we calculate the intersection of this line with the plane...
		// (the camera is assumed to be at (0,0,0))
		double uu = n.dotProduct(p0) / (n.dotProduct(direction));
		
		// calculate the actual point
		FTriple planePoint = direction.times(uu);
		
		// calculate the vector in the plane...
		FTriple planeVector = planePoint.minus(p0);
		
		// project that point onto u and v to get the coordinates...
		double pu = planeVector.dotProduct(u);
		double pv = planeVector.dotProduct(v);
		
		// return these values...
		return new FTriple(pu, pv, 0.0);
	}
	
	public FTriple UVtoXY(FTriple UV)
	{
		if(UV == null)
			return null;
		
		// first, calculate the space coordinates from the UV
		// p = (U * u + V * v) + p0
		FTriple p = ((u.times(UV.val[0])).plus(v.times(UV.val[1]))).plus(p0);
		
		// now transform that into the image plane via:
		// ix = F * (p.x / p.z)
		// iy = F * (p.y / p.z)
		
		// we expect that it should be in the -z direction to be seen
		if(p.val[2] >= 0)
			return null;
			
		double ix = F * (p.val[0] / Math.abs(p.val[2]));
		double iy = F * (p.val[1] / Math.abs(p.val[2]));
		
		// convert to pixel coordinates...
		double px = (ix / halfWidth) * pixelHalfWidth;
		double py = (iy / halfHeight) * pixelHalfHeight;
		
		return new FTriple(px, py, 0.0);
	}
	
	public static void main(String[] args)
	{
		double degangle = 0.0;
		double radangle = (degangle / 180.0) * Math.PI;
		
		FTriple n = new FTriple(0.0, Math.sin(radangle), Math.cos(radangle));
		FTriple u = new FTriple(1.0, 0.0, 0.0);
		FTriple v = (n.crossProduct(u)).norm();
		
		System.out.println("v: " + v);
		
		ImagePlane foo = new ImagePlane(new FTriple(0.0, 0.0, -10.0), n, u, v, 1.0, 4.0/3.0, 240.0);
		FTriple startPoint = new FTriple(100.0, -30.0, 0.0);
		
		FTriple UV = foo.XYtoUV(startPoint);
		
		System.out.println("UV: " + UV);
		
		FTriple XY = foo.UVtoXY(UV);
		
		System.out.println("XY: " + XY);
		
		FTriple UV2 = foo.XYtoUV(XY);
		
		System.out.println("UV2: " + UV2);
	}
}