import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import javax.swing.*;

public class ManualCalibrationFilter extends EffectFilter implements ChangeListener
{	
	private transient JSlider xTranslateSlider;
	private transient JSlider yTranslateSlider;
	private transient JSlider xScaleSlider;
	private transient JSlider yScaleSlider;
	
	protected IntPoint[][] transformTable;
	
	private double translateX;
	private double translateY;
	private double scaleX;
	private double scaleY;
	
	private boolean rebuildTable;

	public ManualCalibrationFilter()
	{
		filterInputs = new FilterLink[1];
		filterInputs[0] = null;
		dead = false;
		filterName = "Manual Calibration Filter " + (int)(Math.random() * 1000);
		
		guiWidth = 100;
		guiHeight = 100;
		guiX = 150;
		guiY = 150;
		
		translateX = 0.0;
		translateY = 0.0;
		scaleX = 1.0;
		scaleY = 1.0;
		
		rebuildTable = true;
		
		rebuildFromSave();
	}
	
	public void prepareForSave()
	{	
		super.prepareForSave();
		getSliderValues();
	}
	
	public void getSliderValues()
	{
		translateX = xTranslateSlider.getValue();
		translateY = yTranslateSlider.getValue();
		
		scaleX = (double)xScaleSlider.getValue() / 100.0;
		scaleY = (double)yScaleSlider.getValue() / 100.0;
	}
	
	public void setSliderValues()
	{
		System.out.println("Set slider values");
		
		xTranslateSlider.setValue((int)translateX);
		yTranslateSlider.setValue((int)translateY);
		
		xScaleSlider.setValue((int)(scaleX * 100.0));
		yScaleSlider.setValue((int)(scaleY * 100.0));
	}
	
	public void stateChanged(ChangeEvent e)
	{
		// indicate that the table should be rebuilt
		rebuildTable = true;
	}
	
	public void rebuildFromSave()
	{
		internalBuffer = new FloatImage();
		needsProcessing = true;
		rebuildTable = true;
		
		guiWindow = new JInternalFrame(filterName, true, true, true, true);
		guiWindow.setSize(guiWidth, guiHeight);
		guiWindow.setLocation(guiX, guiY);
		
		JPanel whatever = new JPanel();
		
		xTranslateSlider = new JSlider(JSlider.HORIZONTAL);
		xTranslateSlider.setMajorTickSpacing(20);
		xTranslateSlider.setMinorTickSpacing(10);
		xTranslateSlider.setPaintTicks(true);
		xTranslateSlider.setBorder(BorderFactory.createTitledBorder("X Translation"));
		xTranslateSlider.setMinimum(-200);
		xTranslateSlider.setMaximum(200);
		xTranslateSlider.addChangeListener(this);
		
		whatever.add(xTranslateSlider);
		
		yTranslateSlider = new JSlider(JSlider.HORIZONTAL);
		yTranslateSlider.setMajorTickSpacing(20);
		yTranslateSlider.setMinorTickSpacing(10);
		yTranslateSlider.setPaintTicks(true);
		yTranslateSlider.setBorder(BorderFactory.createTitledBorder("Y Translation"));
		yTranslateSlider.setMinimum(-200);
		yTranslateSlider.setMaximum(200);
		yTranslateSlider.addChangeListener(this);
		
		whatever.add(yTranslateSlider);
		
		xScaleSlider = new JSlider(JSlider.HORIZONTAL);
		xScaleSlider.setMajorTickSpacing(20);
		xScaleSlider.setMinorTickSpacing(10);
		xScaleSlider.setPaintTicks(true);
		xScaleSlider.setBorder(BorderFactory.createTitledBorder("X Scaling"));
		xScaleSlider.setMinimum(-200);
		xScaleSlider.setMaximum(200);
		xScaleSlider.addChangeListener(this);
		
		whatever.add(xScaleSlider);
		
		yScaleSlider = new JSlider(JSlider.HORIZONTAL);
		yScaleSlider.setMajorTickSpacing(20);
		yScaleSlider.setMinorTickSpacing(10);
		yScaleSlider.setPaintTicks(true);
		yScaleSlider.setBorder(BorderFactory.createTitledBorder("Y Scaling"));
		yScaleSlider.setMinimum(-200);
		yScaleSlider.setMaximum(200);
		yScaleSlider.addChangeListener(this);
		
		whatever.add(yScaleSlider);
		
		setSliderValues();
		
		buildTable();
		
		guiWindow.add(whatever);
		
		guiWindow.setVisible(true);
	}
	
	public String getName()
	{
		return filterName;
	}
	
	public String getInputName(int inputNum)
	{
		return "Input";
	}
	
	public String getOutputName(int outputNum)
	{
		return "Output";
	}
	
	public int numOutputs()
	{
		return 1;
	}
	
	public void buildTable()
	{
		System.out.println("TX: " + translateX);
		System.out.println("TY: " + translateY);
		System.out.println("SX: " + scaleX);
		System.out.println("SY: " + scaleY);	
	
		int imgWidth = internalBuffer.getWidth();
		int imgHeight = internalBuffer.getHeight();
		
		transformTable = new IntPoint[imgWidth][imgHeight];
		
		double cx = imgWidth / 2.0;
		double cy = imgHeight / 2.0;
		
		int ti, tj;
		double dx, dy, tdx, tdy;
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				dx = (double)i;
				dy = (double)j;
				
				tdx = (int)(dx * scaleX);
				tdy = (int)(dy * scaleY);
				
				ti = (int)(translateX + tdx);
				tj = (int)(translateY + tdy);
				
				if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
				{
					transformTable[i][j] = new IntPoint(ti, tj);
				}
			}
		}
		
		System.out.println("Table built");
	}
	
	public FloatImage getOutput(int outputChannel)
	{
		if(rebuildTable == true)
		{
			getSliderValues();
			buildTable();
			rebuildTable = false;
			needsProcessing = true;
		}
		
		if(needsProcessing && filterInputs[0] != null)
		{
			FloatImage temp = filterInputs[0].getOutput();
			
			if(temp != null)
			{	
				if(temp.getWidth() != internalBuffer.getWidth() || temp.getHeight() != internalBuffer.getHeight())
				{
					internalBuffer = new FloatImage(3, temp.getWidth(), temp.getHeight());
					
					// if the input size changes, it invalidates the transform table
					transformTable = null;
				}
				
				int imgWidth = internalBuffer.getWidth();
				int imgHeight = internalBuffer.getHeight();
				
				double[][] ored = internalBuffer.getChannelData(0);
				double[][] ogreen = internalBuffer.getChannelData(1);
				double[][] oblue = internalBuffer.getChannelData(2);
				
				double[][] ired = temp.getChannelData(0);
				double[][] igreen = temp.getChannelData(1);
				double[][] iblue = temp.getChannelData(2);
				
				if(transformTable != null)
				{
					int ti, tj;
					
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							if(transformTable[i][j] != null)
							{
								ti = transformTable[i][j].x;
								tj = transformTable[i][j].y;
								
								if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
								{
									//System.out.println("T");
									
									ored[i][j] = ired[ti][tj];
									ogreen[i][j] = igreen[ti][tj];
									oblue[i][j] = iblue[ti][tj];
								}
								else
								{
									ored[i][j] = -1.0;
									ogreen[i][j] = -1.0;
									oblue[i][j] = -1.0;
								}
							}
							else
							{
								ored[i][j] = -1.0;
								ogreen[i][j] = -1.0;
								oblue[i][j] = -1.0;
							}
						}
					}
				}
				else // transformTable == null
				{
					System.out.println("Null tt");
					
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							ored[i][j] = ired[i][j];
							ogreen[i][j] = igreen[i][j];
							oblue[i][j] = iblue[i][j];
						}
					}
				}
			}
			needsProcessing = false;
		}

		return internalBuffer;
	}	
}