import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import javax.swing.*;

public class ManualRotationFilter extends EffectFilter implements ChangeListener
{	
	private transient JSlider thetaSlider;
	
	protected IntPoint[][] transformTable;
	
	private double rotationAmount;
	
	private boolean rebuildTable;
	
	private static final double GRANULARITY = 1000.0;

	public ManualRotationFilter()
	{
		filterInputs = new FilterLink[1];
		filterInputs[0] = null;
		dead = false;
		filterName = "Manual Rotation Filter " + (int)(Math.random() * 1000);
		
		guiWidth = 100;
		guiHeight = 100;
		guiX = 150;
		guiY = 150;
		
		rotationAmount = 0.0;
		
		rebuildTable = true;
		
		rebuildFromSave();
	}
	
	public void prepareForSave()
	{	
		super.prepareForSave();
		getSliderValues();
	}
	
	public void getSliderValues()
	{
		rotationAmount = (double)thetaSlider.getValue() / GRANULARITY;
	}
	
	public void setSliderValues()
	{
		System.out.println("Set slider values");
		
		thetaSlider.setValue((int)(rotationAmount * GRANULARITY));
	}
	
	public void stateChanged(ChangeEvent e)
	{
		// indicate that the table should be rebuilt
		rebuildTable = true;
	}
	
	public void rebuildFromSave()
	{
		internalBuffer = new FloatImage();
		needsProcessing = true;
		rebuildTable = true;
		
		guiWindow = new JInternalFrame(filterName, true, true, true, true);
		guiWindow.setSize(guiWidth, guiHeight);
		guiWindow.setLocation(guiX, guiY);
		
		JPanel whatever = new JPanel();
		
		thetaSlider = new JSlider(JSlider.HORIZONTAL);
		thetaSlider.setMajorTickSpacing(200);
		thetaSlider.setMinorTickSpacing(100);
		thetaSlider.setPaintTicks(true);
		thetaSlider.setBorder(BorderFactory.createTitledBorder("X Translation"));
		thetaSlider.setMinimum(-2 * (int)GRANULARITY);
		thetaSlider.setMaximum(2 * (int)GRANULARITY);
		thetaSlider.addChangeListener(this);
		
		whatever.add(thetaSlider);
		
		setSliderValues();
		
		buildTable();
		
		guiWindow.add(whatever);
		
		guiWindow.setVisible(true);
	}
	
	public String getName()
	{
		return filterName;
	}
	
	public String getInputName(int inputNum)
	{
		return "Input";
	}
	
	public String getOutputName(int outputNum)
	{
		return "Output";
	}
	
	public int numOutputs()
	{
		return 1;
	}
	
	public void buildTable()
	{
		System.out.println("Theta: " + rotationAmount);	
	
		int imgWidth = internalBuffer.getWidth();
		int imgHeight = internalBuffer.getHeight();
		
		transformTable = new IntPoint[imgWidth][imgHeight];
		
		double cx = imgWidth / 2.0;
		double cy = imgHeight / 2.0;
		
		double theta = Math.PI * rotationAmount;
		
		int ti, tj;
		double dx, dy, tdx, tdy;
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				dx = (double)i - cx;
				dy = (double)j - cy;
				
				tdx = (dx * Math.cos(theta) - dy * Math.sin(theta));
				tdy = (dx * Math.sin(theta) + dy * Math.cos(theta));
				
				ti = (int)(cx + tdx);
				tj = (int)(cy + tdy);
				
				if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
				{
					transformTable[i][j] = new IntPoint(ti, tj);
				}
			}
		}
		
		System.out.println("Table built");
	}
	
	public FloatImage getOutput(int outputChannel)
	{
		if(rebuildTable == true)
		{
			getSliderValues();
			buildTable();
			rebuildTable = false;
			needsProcessing = true;
		}
		
		if(needsProcessing && filterInputs[0] != null)
		{
			FloatImage temp = filterInputs[0].getOutput();
			
			if(temp != null)
			{	
				if(temp.getWidth() != internalBuffer.getWidth() || temp.getHeight() != internalBuffer.getHeight())
				{
					internalBuffer = new FloatImage(3, temp.getWidth(), temp.getHeight());
					
					// if the input size changes, it invalidates the transform table
					transformTable = null;
				}
				
				int imgWidth = internalBuffer.getWidth();
				int imgHeight = internalBuffer.getHeight();
				
				double[][] ored = internalBuffer.getChannelData(0);
				double[][] ogreen = internalBuffer.getChannelData(1);
				double[][] oblue = internalBuffer.getChannelData(2);
				
				double[][] ired = temp.getChannelData(0);
				double[][] igreen = temp.getChannelData(1);
				double[][] iblue = temp.getChannelData(2);
				
				if(transformTable != null)
				{
					int ti, tj;
					
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							if(transformTable[i][j] != null)
							{
								ti = transformTable[i][j].x;
								tj = transformTable[i][j].y;
								
								if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
								{
									//System.out.println("T");
									
									ored[i][j] = ired[ti][tj];
									ogreen[i][j] = igreen[ti][tj];
									oblue[i][j] = iblue[ti][tj];
								}
								else
								{
									ored[i][j] = -1.0;
									ogreen[i][j] = -1.0;
									oblue[i][j] = -1.0;
								}
							}
							else
							{
								ored[i][j] = -1.0;
								ogreen[i][j] = -1.0;
								oblue[i][j] = -1.0;
							}
						}
					}
				}
				else // transformTable == null
				{
					System.out.println("Null tt");
					
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							ored[i][j] = ired[i][j];
							ogreen[i][j] = igreen[i][j];
							oblue[i][j] = iblue[i][j];
						}
					}
				}
			}
			needsProcessing = false;
		}

		return internalBuffer;
	}	
}