

import java.awt.Color;
import java.awt.Graphics;
import java.util.Vector;
import javax.swing.*;
import java.awt.geom.*;

public class MonteCarlo extends JComponent {
	private static final double BEARING_UNCERTAINTY = 0.1;  // uncertainty on direction of camera
	private static final double POST_UNCERTAINTY = 0.1;     // uncertainty on post identification
	private static final double ANGLE_UNCERTAINTY = 1.0;    // uncertainty on absolute post angle
	private static final double DISTANCE_UNCERTAINTY = 0.05; // uncertainty on distance to post
	
	public static Vector<Particle> particles;
	private static Particle mostLikely;
	
	/**
	 * create a monte carlo simulation with numParticles particles
	 */
	public static void createParticles(int numParticles) {
		particles = new Vector<Particle>();
		for(int i=0; i<numParticles; i++) {
			particles.add(new Particle());
		}
		mostLikely = particles.get(0);
	}
	
	/**
	 * return a copy of the most likely particle
	 */
	public static Particle getMostLikely() {
		return new Particle(mostLikely);
	}
	
	/**
	 * update particles based on odometry from the robot
	 */
	public static void updateParticles(double deltaR, double deltaTheta) {
		MotionModel m = new MotionModel(deltaR, deltaTheta);
		
		for(Particle p : particles) {
			m.applyMotion(p);
		}
	}
	
	/**
	 * cull the particles based on an observed post
	 */
	public static void postSeen(int postNum, double bearing, double absAngle, double distance) {
		double totalWeights = 0;
		mostLikely = null;
		
		// calculate the weights
		for(Particle p : particles) {
			p.probability = 0;
			for(int pNum : PlayingField.posts) {
				double angleToPost = PlayingField.angleToPost(p, pNum);
				double angleOnPost = PlayingField.angleOnPost(p, pNum);
				double distanceToPost = PlayingField.distanceToPost(p, pNum);
				double newProb = 0;
				// probability of seeing post where we saw it
				newProb += 1/(BEARING_UNCERTAINTY*(BEARING_UNCERTAINTY + Math.abs(Math.sin((bearing - angleToPost)/2))));
				// probability of seeing absolute angle on post
				newProb += 1/(ANGLE_UNCERTAINTY*(ANGLE_UNCERTAINTY + Math.abs(Math.sin((absAngle - angleOnPost)/2))));
				// probability of seeing the distance of the post
				newProb += 1/(DISTANCE_UNCERTAINTY*(DISTANCE_UNCERTAINTY + PlayingField.normalizeDistance(Math.abs(distance - distanceToPost))));
				
				// probability of seeing the post we think we saw
				if(pNum != postNum)
					newProb *= POST_UNCERTAINTY;
				
				p.probability += Math.pow(newProb,2);
				//p.theta = angleToPost;
				//p.probability = 1;
			}
			
			// particles outside the boundary aren't very likely
			if(!PlayingField.isInBounds(p))
				p.probability *= 0.1;
			
			if(mostLikely == null || p.probability > mostLikely.probability)
				mostLikely = new Particle(p);
			
			totalWeights += p.probability;
		}
		
		// reweight the probabilities to averate to 1
		double w = particles.size()/totalWeights;
		for(Particle p : particles) {
			p.probability *= w;
		}
		
		// cull low probabilty particles
		int particlesLeft = particles.size();
		Vector<Particle> newParticles = new Vector<Particle>();
		
		for(Particle p : particles){
			// exceptionally likely particles will be automatically included
			int foSho = (int)p.probability;
			for(int i=0; i<foSho; ++i) {
				newParticles.add(new Particle(p));
			}
			particlesLeft -= foSho;
			p.probability -= foSho;
		}
		
		for(int i=0; i<particlesLeft; ++i) {
			// fill in the remaining spaces with proportionally chosen random particles
			double r = particlesLeft * PlayingField.randUniform();
			for(Particle p : particles) {
				r -= p.probability;
				if(r < 0) {
					newParticles.add(new Particle(p));
					break;
				}
			}
		}
		
		// out with the old, in with the new
		particles = newParticles;
	}
	
	
	public void paint(Graphics g) {
		Rectangle2D.Double bound = PlayingField.getBoundary();
		
		for(Particle p : particles) {
			if(p.probability > 0) {
				int x = (int)p.x;
				int y = (int)p.y;
				g.setColor(new Color(Math.min(128*(int)p.probability, 255),0,0));
				g.fillOval(x, y, 6, 6);
				g.setColor(Color.black);
				g.drawLine(x+3, y+3, (int)(x+3+6*Math.cos(p.theta)), (int)(y+3+6*Math.sin(p.theta)));
			}
		}
		g.setColor(Color.blue);
		g.fillOval((int)mostLikely.x, (int)mostLikely.y, 10, 10);

		g.setColor(Color.green);
		g.drawLine((int)mostLikely.x+5, (int)mostLikely.y+5, (int)(mostLikely.x+5+10*Math.cos(mostLikely.theta)), (int)(mostLikely.y+5+ 10*Math.sin(mostLikely.theta)));

		
		g.setColor(Color.black);
		g.drawRect(0,0,(int)bound.width,(int)bound.height);
	}
}
