import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;

public class ProjectorTransformFilter extends EffectFilter
{	
	protected IntPoint[][] transformTable;
	protected transient Calibrator theCalibrator;
	protected transient int status;
	protected int pauseFrames;
	
	private static final int NUM_PAUSE = 20;
	
	private static final int STATUS_TRANSFORM = 23;
	private static final int STATUS_CALIBRATE = 67;

	public ProjectorTransformFilter()
	{
		filterInputs = new FilterLink[2];
		filterInputs[0] = null;
		filterInputs[1] = null;
		dead = false;
		filterName = "Projector Transform " + (int)(Math.random() * 1000);
		
		guiWidth = 300;
		guiHeight = 300;
		guiX = 200;
		guiY = 200;
		
		transformTable = null;
		
		pauseFrames = 0;
		
		rebuildFromSave();
	}
	
	public void rebuildFromSave()
	{
		internalBuffer = new FloatImage();
		theCalibrator = null;
		status = STATUS_TRANSFORM;
			
		needsProcessing = true;
		
		guiWindow = new JInternalFrame(filterName, true, true, true, true);
		guiWindow.setSize(guiWidth, guiHeight);
		guiWindow.setLocation(guiX, guiY);
		
		JPanel foo = new JPanel();
		JButton loadButton = new JButton("Calibrate");
		loadButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e)
			{
				// must be a button press
				startCalibration();
			}
		});
		foo.add(loadButton);

		//theScreen = new FloatImageScreen();
		//theScreen.setPreferredSize(new Dimension(160,120));

		
		//foo.add(theScreen);
		
		guiWindow.add(foo);
		//guiWindow.add(new JScrollPane(theScreen))
		
		guiWindow.setVisible(true);
	}
	
	public void prepareForSave()
	{
		guiWidth = guiWindow.getWidth();
		guiHeight = guiWindow.getHeight();
		guiX = guiWindow.getX();
		guiY = guiWindow.getY();
	}
	
	public void startCalibration()
	{
		// combo box of possible calibration methods
		String[] cmethods = new String[]{"Gradient", "Sweep", "Binary", "TranslateScale", "Click TS"};
		JComboBox inputSelection = new JComboBox(cmethods);
		inputSelection.setEditable(false);
		
		// Now, we show a dialog box to let the user select from these
		int result = JOptionPane.showOptionDialog(guiWindow, 
					new Object[]{inputSelection},
					"Select Calibration Method", JOptionPane.OK_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE, null, null, null);
					
		// If the user clicked 'OK', then we attempt to make the connection
		int selectedMethod = inputSelection.getSelectedIndex();
		
		if(result == JOptionPane.OK_OPTION && selectedMethod != -1)
		{
			System.out.println("Attempting to start calibration");
			internalBuffer = new FloatImage();
			switch(selectedMethod)
			{
			case 0: // gradient
				System.out.println("Attempting gradient calibration");
				break;
			case 1: // sweep
				System.out.println("Attempting sweep calibration");
				theCalibrator = new SweepCalibration();
				break;
			case 2: // binary
				System.out.println("Attempting binary calibration");
				theCalibrator = new BinaryCalibration();
				break;
			case 3: // translateScale
				System.out.println("Attempting translate scale calibration");
				theCalibrator = new TranslateScaleCalibration();
				break;
			case 4: // manual translateScale
				System.out.println("Attempting translate scale calibration with clicking");
				theCalibrator = new TranslateScaleCalibration(guiWindow);
				break;
			}
		}
		
		if(theCalibrator != null)
		{
			status = STATUS_CALIBRATE;
			pauseFrames = NUM_PAUSE;
		}
	}
	
	public String getOutputName(int outputNum)
	{
		return "ImageOut";
	}
	
	public String getInputName(int inputNum)
	{
		if(inputNum == 0)
			return "ImageIn";
		else
			return "CalibrationInput";
	}
	
	public int numOutputs()
	{
		return 1;
	}
	
	public void update(long dt)
	{
		needsProcessing = true;
		// do nothing...
	}
	
	public FloatImage getOutput(int outputChannel)
	{
		if(status != STATUS_CALIBRATE)
		{
			return doTransform();
		}
		else // status == STATUS_CALIBRATE
		{
			if(pauseFrames > 0)
			{
				pauseFrames--;
				return internalBuffer;
			}
			else
			{
				pauseFrames = NUM_PAUSE;
				return doCalibrate();
			}
		}
		
	}
	
	private FloatImage doCalibrate()
	{
		if(theCalibrator == null)
		{
			status = STATUS_TRANSFORM;
			return internalBuffer;
		}
		
		if(filterInputs[1] == null)
			return internalBuffer;
			
		FloatImage temp = filterInputs[1].getOutput();
		
		if(temp == null)
			return internalBuffer;
			
		int tempWidth = temp.getWidth();
		int tempHeight = temp.getHeight();
			
		theCalibrator.processResults(temp);
		
		temp = theCalibrator.getNextImage();
		
		internalBuffer = temp;
		
		if(temp == null)
		{
			System.out.println("Done calibrating");
			transformTable = theCalibrator.getTransform();
			if(transformTable == null)
				System.out.println("Warning! Null transform!");
			status = STATUS_TRANSFORM;
			
			internalBuffer = new FloatImage(3, tempWidth, tempHeight);
			
			return internalBuffer;	
		}
		else
		{
			return temp;
		}
	}
	
	private FloatImage doTransform()
	{
		if(needsProcessing && filterInputs[0] != null)
		{
			FloatImage temp = filterInputs[0].getOutput();
			
			if(temp != null)
			{	
				if(temp.getWidth() != internalBuffer.getWidth() || temp.getHeight() != internalBuffer.getHeight())
				{
					internalBuffer = new FloatImage(3, temp.getWidth(), temp.getHeight());
					
					// if the input size changes, it invalidates the transform table
					transformTable = null;
				}
				
				int imgWidth = internalBuffer.getWidth();
				int imgHeight = internalBuffer.getHeight();
				
				double[][] ored = internalBuffer.getChannelData(0);
				double[][] ogreen = internalBuffer.getChannelData(1);
				double[][] oblue = internalBuffer.getChannelData(2);
				
				double[][] ired = temp.getChannelData(0);
				double[][] igreen = temp.getChannelData(1);
				double[][] iblue = temp.getChannelData(2);
				
				if(transformTable != null)
				{
					int ti, tj;
					
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							if(transformTable[i][j] != null)
							{
								ti = transformTable[i][j].x;
								tj = transformTable[i][j].y;
								
								if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
								{
									ored[i][j] = ired[ti][tj];
									ogreen[i][j] = igreen[ti][tj];
									oblue[i][j] = iblue[ti][tj];
								}
							}
						}
					}
				}
				else // transformTable == null
				{
					for(int i = 0; i < imgWidth; ++i)
					{
						for(int j = 0; j < imgHeight; ++j)
						{
							ored[i][j] = ired[i][j];
							ogreen[i][j] = igreen[i][j];
							oblue[i][j] = iblue[i][j];
						}
					}
				}
			}
			needsProcessing = false;
		}

		return internalBuffer;
	}	
}