public class SimpleCycler
{
	private double[][] cyclingField;
	private double[][] coolingField;
	private double[][] hotSpots;
	private int fieldWidth;
	private int fieldHeight;
	private FloatImage internalBuffer;
	
	private static final double DEFAULT_CYCLING = 0.02;
	private static final double CYCLING_MULT = 0.01;
	
	public SimpleCycler()
	{
		fieldWidth = 100;
		fieldHeight = 100;
		
		cyclingField = new double[fieldWidth][fieldHeight];
		createUniformCycling(DEFAULT_CYCLING);
		createDefaultHotspots();
	}

	public void setCyclingField(FloatImage src)
	{
		if(src == null)
		{
			createUniformCycling(DEFAULT_CYCLING);
			return;
		}
		
		int iwidth = Math.min(fieldWidth, src.getWidth());
		int iheight = Math.min(fieldHeight, src.getHeight());
		
		double[][] red = src.getChannelData(0);
		double[][] green = src.getChannelData(1);
		double[][] blue = src.getChannelData(2);
		
		cyclingField = new double[fieldWidth][fieldHeight];
		
		for(int i = 0; i < iwidth; ++i)
		{
			for(int j = 0; j < iheight; ++j)
			{
				coolingField[i][j] = (((red[i][j] + blue[i][j] + green[i][j]) / 3.0) + 1.0) * CYCLING_MULT;	
			}
		}
	}
	
	public void setHotSpots(double[][] hots)
	{
		if(hots == null)
			return;
			
		int nheight = hots[0].length;
		int nwidth = hots.length;
		
		if(nheight != fieldHeight || nwidth != fieldWidth)
		{
			fieldHeight = nheight;
			fieldWidth = nwidth;
			cyclingField = new double[fieldWidth][fieldHeight];
			createUniformCycling(DEFAULT_CYCLING);
		}
		
		hotSpots = hots;
	}
	
	public void createDefaultHotspots()
	{
		hotSpots = new double[fieldWidth][fieldHeight];
	}
	
	public void createUniformCycling(double amt)
	{
		coolingField = new double[fieldWidth][fieldHeight];
		
		for(int i = 0; i < fieldWidth; ++i)
		{
			for(int j = 0; j < fieldHeight; ++j)
			{
				coolingField[i][j] = amt;
			}
		}
	}

	public void updateCycling()
	{
		double tempVal;
		
		for(int j = 0; j < fieldHeight; ++j)
		{
			for(int i = 0; i < fieldWidth; ++i)
			{
				// just add in the hotspots and subtract out the cooling
				tempVal = cyclingField[i][j];
				
				tempVal = tempVal + hotSpots[i][j] + 1.0 - coolingField[i][j];
				
				if(tempVal > 1.0)
					tempVal = 1.0;
					
				if(tempVal < 0.0)
					tempVal = 0.0;
					
				/*
				if(new Double(tempVal).isNaN())
					System.out.println("Not a number!");
				*/	
					
				cyclingField[i][j] = tempVal;
			}
		}
	}
	
	public FloatImage getCyclingImage(FloatImage cyclingColors)
	{
		if(internalBuffer == null || internalBuffer.getWidth() != fieldWidth 
			|| internalBuffer.getHeight() != fieldHeight)
		{
			internalBuffer = new FloatImage(3, fieldWidth, fieldHeight);
		}
		
		if(cyclingColors == null)
			return internalBuffer;
		
		double[][] ored = internalBuffer.getChannelData(0);
		double[][] ogreen = internalBuffer.getChannelData(1);
		double[][] oblue = internalBuffer.getChannelData(2);
		
		double[][] pred = cyclingColors.getChannelData(0);
		double[][] pgreen = cyclingColors.getChannelData(1);
		double[][] pblue = cyclingColors.getChannelData(2);
		
		int numBins = cyclingColors.getWidth() - 1;
		int colorIndex;
		
		for(int i = 0; i < fieldWidth; ++i)
		{
			for(int j = 0; j < fieldHeight; ++j)
			{
				colorIndex = (int)((1.0 - cyclingField[i][j]) * numBins);
				ored[i][j] = pred[colorIndex][0];
				ogreen[i][j] = pgreen[colorIndex][0];
				oblue[i][j] = pblue[colorIndex][0];
			}
		}
		
		return internalBuffer;
	}

}