import java.awt.*;
import java.awt.image.*;
import javax.swing.*;

public class SimpleDrawerRobot implements GenericRobot
{
	protected final static int MODE_ROTATE = 1;
	protected final static int MODE_GO =2;
	protected final static int MODE_STOPPED = 3;
	
	protected final static double ROTATE_SPEED_FACTOR = 0.002;
	protected final static double MIN_ROTATE_SPEED = 0.1;
	protected final static double THETA_LIMIT = 0.5;
	protected final static double MAX_THETA_ERROR = 0.5;
	protected final static double MAX_SPEED = 0.5;
	protected final static double MIN_SPEED = 0.1;
	protected final static double GO_SPEED_FACTOR = 0.003;
	protected final static double TARGET_DISTANCE = 1.0;
	
	long lastTime;
	
	protected int mode;
	
	protected double targetX;
	protected double targetY;
	
	protected JFrame theFrame;
	protected RobotDisplay roboDisplay;
	
	public void update(long dt, FilterLink[] inputs)
	{
		System.out.println("Update!");
		
		RoombaControl theControl = RoombaControl.getInstance();
		
		long curTime = System.currentTimeMillis();
		dt = curTime - lastTime;
		lastTime = curTime;
		
		if(dt < 0 || dt > 1000)
			dt = 20;
			
		// first, get the location and angle
		double curX = -1000.0;
		double curY = -1000.0;
		
		double curTheta = 0.0;
		
		if(inputs[0] != null)
		{
			FloatImage p1i = inputs[0].getOutput();
			
			if(p1i != null)
			{
				double p1x, p2x, p1y, p2y;
				double[][] temput = p1i.getChannelData(0);
				p1x = temput[0][0];
				p1y = temput[1][0];
				p2x = temput[0][1];
				p2y = temput[1][1];
				
				// we say the current position is the average of these two!
				curX = (p1x + p2x) / 2.0;
				curY = (p1y + p2y) / 2.0;
				
				// compute theta!
				// we have to offset by 90 degrees since the dots are perpendicular
				// to the direction the robot is actually facing...
				curTheta = Math.atan2(p2y - p1y, p2x - p1x) - (Math.PI / 2.0);
				while(curTheta < 0.0)
					curTheta += (Math.PI * 2.0);
			}
			else
			{
				System.out.println("Null input!");
				return;
			}
		}
		else
		{
			System.out.println("Null image!");
			return;
		}
		

		// update robot display!
		if(roboDisplay == null)
		{
			System.out.println("Null display!");
			return;
		}
			
		roboDisplay.setRobotPos(curX, curY, curTheta);
		roboDisplay.repaint();
		
		// grab target!
		targetX = roboDisplay.getTargetX();
		targetY = roboDisplay.getTargetY();
		
		double targetTheta = Math.atan2(targetY - curY, targetX - curX);
		if(targetTheta < 0.0)
			targetTheta += (Math.PI * 2.0);
			
		double dx = targetX - curX;
		double dy = targetY - curY;
		double distance = Math.sqrt((dx*dx) + (dy*dy));
		
		if(distance < TARGET_DISTANCE)
		{
			mode = MODE_STOPPED;
		}
		
		if(mode == MODE_STOPPED)
		{
			System.out.println("Stopped Mode");	
			
			theControl.directDrive(0.0, 0.0);
		
			if(distance > TARGET_DISTANCE)
			{
				mode = MODE_ROTATE;
				return;
			}
		}
		else if(mode == MODE_GO)
		{
			System.out.println("Go Mode");
			
			if(Math.abs(targetTheta - curTheta) > MAX_THETA_ERROR)
			{
				mode = MODE_ROTATE;
				return;
			}
			
			double goSpeed = distance * GO_SPEED_FACTOR;
			goSpeed = Math.min(goSpeed, MAX_SPEED);
			goSpeed = Math.max(goSpeed, MIN_SPEED);
			
			System.out.println("Going with speed " + goSpeed);
			
			theControl.directDrive(goSpeed, goSpeed);
		}
		else if(mode == MODE_ROTATE)
		{
			System.out.println("Rotate Mode");
			
			if(Math.abs(targetTheta - curTheta) < THETA_LIMIT)
			{
				mode = MODE_GO;
				return;
			}
			
			double diff = targetTheta - curTheta;
			double rSpeed = Math.abs(diff) * ROTATE_SPEED_FACTOR;
			rSpeed = Math.max(rSpeed, MIN_ROTATE_SPEED);
			rSpeed = (diff / Math.abs(diff)) * rSpeed;
			
			System.out.println("Rotating with speed " + rSpeed);
			
			theControl.directDrive(rSpeed, -rSpeed);
		}
	}
	
	public void init()
	{
		RoombaControl theControl = RoombaControl.getInstance();
		
		theControl.startOI();
		theControl.setSafeMode();
		
		mode = MODE_STOPPED;
		
		JFrame theFrame = new JFrame("Frame!");
		theFrame.setSize(300,300);
		
		roboDisplay = new RobotDisplay();
		
		theFrame.getContentPane().add(roboDisplay);
		theFrame.setVisible(true);
	}
}