public class SimpleFire
{
	private double[][] fireField;
	private double[][] coolingField;
	private double[][] hotSpots;
	private int fieldWidth;
	private int fieldHeight;
	private int frameCounter;
	private FloatImage internalBuffer;
	
	private static final double DEFAULT_COOLING = 0.2;
	
	public SimpleFire()
	{
		frameCounter = 0;
		fieldWidth = 100;
		fieldHeight = 100;
		
		fireField = new double[fieldWidth][fieldHeight];
		createUniformCooling(DEFAULT_COOLING);
		createDefaultFireHotspots();
	}

	public void setCoolingField(FloatImage src)
	{
		if(src == null)
		{
			createUniformCooling(DEFAULT_COOLING);
			return;
		}
		
		int iwidth = Math.min(fieldWidth, src.getWidth());
		int iheight = Math.min(fieldHeight, src.getHeight());
		
		double[][] red = src.getChannelData(0);
		double[][] green = src.getChannelData(1);
		double[][] blue = src.getChannelData(2);
		
		coolingField = new double[fieldWidth][fieldHeight];
		
		for(int i = 0; i < iwidth; ++i)
		{
			for(int j = 0; j < iheight; ++j)
			{
				coolingField[i][j] = ((red[i][j] + blue[i][j] + green[i][j]) / 3.0) + 1.0;	
			}
		}
	}
	
	public void setHotSpots(double[][] hots)
	{
		if(hots == null)
			return;
			
		int nheight = hots[0].length;
		int nwidth = hots.length;
		
		if(nheight != fieldHeight || nwidth != fieldWidth)
		{
			fieldHeight = nheight;
			fieldWidth = nwidth;
			fireField = new double[fieldWidth][fieldHeight];
			createUniformCooling(DEFAULT_COOLING);
		}
		
		hotSpots = hots;
	}
	
	public void createDefaultFireHotspots()
	{
		hotSpots = new double[fieldWidth][fieldHeight];
		for(int i = 0; i < fieldWidth; ++i)
		{
			hotSpots[i][fieldHeight-1] = Math.random();
		}	
	}
	
	public void createUniformCooling(double amt)
	{
		coolingField = new double[fieldWidth][fieldHeight];
		
		for(int i = 0; i < fieldWidth; ++i)
		{
			for(int j = 0; j < fieldHeight; ++j)
			{
				coolingField[i][j] = amt;
			}
		}
	}

	public void updateFire()
	{
		double tempSum;
		
		for(int j = 0; j < fieldHeight - 1; ++j)
		{
			for(int i = 2; i < fieldWidth - 2; ++i)
			{
				// do some blending/averaging
				tempSum = (fireField[i-2][j+1] + fireField[i-1][j+1] + 
							fireField[i][j+1] + fireField[i+1][j+1] + 
							fireField[i+2][j+1]) / 5.0;
				
				// do some hot spot stuff
				tempSum += (hotSpots[i][j] + 1.0);
				
				// do some cooling stuff
				tempSum -= (coolingField[i][(j+frameCounter) % fieldHeight]) * 0.2;
				
				// make sure it is still within range
				
				if(tempSum > 1.0)
					tempSum = 1.0;
				else if(tempSum < 0.0)
					tempSum = 0.0;
					
				// finally, the fireField at this location becomes the temporary
				//  value
				fireField[i][j] = tempSum;
			}
		}
		
		// update frame counter
		frameCounter = (frameCounter + 1) % fieldHeight;
	}
	
	public FloatImage getFireImage(FloatImage fireColors)
	{
		if(internalBuffer == null || internalBuffer.getWidth() != fieldWidth 
			|| internalBuffer.getHeight() != fieldHeight)
		{
			internalBuffer = new FloatImage(3, fieldWidth, fieldHeight);
		}
		
		if(fireColors == null)
			return internalBuffer;
		
		double[][] ored = internalBuffer.getChannelData(0);
		double[][] ogreen = internalBuffer.getChannelData(1);
		double[][] oblue = internalBuffer.getChannelData(2);
		
		double[][] pred = fireColors.getChannelData(0);
		double[][] pgreen = fireColors.getChannelData(1);
		double[][] pblue = fireColors.getChannelData(2);
		
		int numBins = fireColors.getWidth() - 1;
		int colorIndex;
		
		for(int i = 0; i < fieldWidth; ++i)
		{
			for(int j = 0; j < fieldHeight; ++j)
			{
				colorIndex = (int)((1.0 - fireField[i][j]) * numBins);
				ored[i][j] = pred[colorIndex][0];
				ogreen[i][j] = pgreen[colorIndex][0];
				oblue[i][j] = pblue[colorIndex][0];
			}
		}
		
		return internalBuffer;
	}

}