import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;

public class TranslateScaleCalibration implements Calibrator, MouseListener
{
	private IntPoint[][] transformTable;
	private double[][] baseline;
	private FloatImage internalBuffer;
	
	private Component parentComponent;
	private int identifyMode;
	private int captureX;
	private int captureY;
	
	private static final int IDENTIFY_ALG = 1;
	private static final int IDENTIFY_CLK = 2;
	
	private IntPoint firstDotProj;
	private IntPoint firstDotCam;
	private IntPoint secondDotProj;
	private IntPoint secondDotCam;
	
	private double scaleFactor;
	
	private int imgWidth;
	private int imgHeight;
	
	private static final double THRESHOLD = 0.5;
	private static final int DOT_SIZE = 6;
	
	private int mode;
	
	private static final int MODE_INITIALIZE = 1;
	private static final int MODE_FIRSTDOT = 2;
	private static final int MODE_SECONDDOT = 3;
	private static final int MODE_IDENTIFY = 4;
	private static final int MODE_DONE = 5;
	private static final int MODE_ALMOSTDONE = 6;
	
	public TranslateScaleCalibration()
	{
		transformTable = null;
		baseline = null;
		
		imgWidth = 0;
		imgHeight = 0;
		
		internalBuffer = null;
		
		captureX = 0;
		captureY = 0;
		
		mode = MODE_INITIALIZE;
		identifyMode = IDENTIFY_ALG;
		parentComponent = null;
	}
	
	// sets up TranslateScale to ask to user to click on the dots rather than
	//  attempting to identify them itself
	
	public TranslateScaleCalibration(Component p)
	{
		transformTable = null;
		baseline = null;
		
		imgWidth = 0;
		imgHeight = 0;
		
		internalBuffer = null;
		
		mode = MODE_INITIALIZE;
		identifyMode = IDENTIFY_CLK;
		parentComponent = p;
		
		captureX = 0;
		captureY = 0;
	}	
	
	public FloatImage getNextImage()
	{
		switch(mode)
		{
		case MODE_INITIALIZE:
			makeAllBlack();
			mode = MODE_FIRSTDOT;
			break;
		case MODE_FIRSTDOT:
			firstDotProj = makeDot();
			mode = MODE_SECONDDOT;
			break;
		case MODE_SECONDDOT:
			secondDotProj = makeDot();
			mode = MODE_ALMOSTDONE;
			break;
		case MODE_DONE:
			return null;
		}
		
		return internalBuffer;
	}
	
	public void processResults(FloatImage result)
	{
		if(result != null)
		{
			switch(mode)
			{
			case MODE_INITIALIZE:
				imgWidth = result.getWidth();
				imgHeight = result.getHeight();
				internalBuffer = new FloatImage(3, imgWidth, imgHeight);
				break;
			case MODE_FIRSTDOT:
				calculateBaseline(result);
				break;
			case MODE_SECONDDOT:
				firstDotCam = identifyDot(result);
				break;
			case MODE_ALMOSTDONE:
				secondDotCam = identifyDot(result);
				mode = MODE_DONE;
				break;
			}
		}
	}
	
	public IntPoint makeDot()
	{	
		// First, fill with black
	
		double[][] red = internalBuffer.getChannelData(0);
		double[][] green = internalBuffer.getChannelData(1);
		double[][] blue = internalBuffer.getChannelData(2);
		
		double fillValue = -1.0;
		
		for(int j = 0; j < imgHeight; ++j)
		{	
			for(int i = 0; i < imgWidth; ++i)
			{
				red[i][j] = fillValue;
				green[i][j] = fillValue;
				blue[i][j] = fillValue;
			}
		}
		
		// Now, generate the Dot's coordinates:
		int dotx = (int)(Math.random() * (imgWidth - DOT_SIZE)); 
		int doty = (int)(Math.random() * (imgHeight - DOT_SIZE));
		
		for(int i = 0; i < DOT_SIZE; ++i)
		{
			for(int j = 0; j < DOT_SIZE; ++j)
			{
				red[dotx + i][doty + j] = 1.0;
				green[dotx + i][doty + j] = 1.0;
				blue[dotx + i][doty + j] = 1.0;
			}
		}
		
		return new IntPoint(dotx, doty);
	}
	
	private double[][] blur(double[][] src)
	{
		double[][] result = new double[imgWidth][imgHeight];
		
		double sum = 0.0;
		
		for(int i = 1; i < imgWidth - 1; ++i)
		{
			for(int j = 1; j < imgHeight - 1; ++j)
			{
				sum = src[i-1][j-1] + src[i][j-1] + src[i+1][j-1] +
						 src[i-1][j] + src[i][j] + src[i+1][j] +
						 src[i-1][j+1] + src[i][j+1] + src[i+1][j+1];
						 
				result[i][j] = sum / 9.0;
			}
		}
		
		return result;
		
	}
	
	public IntPoint identifyDot(FloatImage src)
	{
		if(identifyMode == IDENTIFY_ALG)
		{	
			double[][] gain = makeGain(src);
			
			// blur the gain a bit
			gain = blur(gain);
			gain = blur(gain);
			
			double bestGain = 0.0;
			int bestX = 0;
			int bestY = 0;
			
			// this is because the blurring does not blur the very edges
			for(int i = 1; i < imgWidth-1; ++i)
			{
				for(int j = 1; j < imgHeight-1; ++j)
				{		
					if(gain[i][j] > bestGain)
					{
						bestX = i;
						bestY = j;
						bestGain = gain[i][j];
					}
				}
			}
			
			return new IntPoint(bestX, bestY);
		}
		else
		{
			FloatImageScreen screen = new FloatImageScreen();
			screen.imgUpdate(src);
			screen.setMode(FloatImageScreen.MODE_FIXED);
			String prompt = "Click on the dot";
			screen.setPreferredSize(new Dimension(src.getWidth(), src.getHeight()));
			System.out.println("Preferred Dimensions: " + src.getWidth() + ", " + src.getHeight());
			screen.addMouseListener(this);
			
			int result = JOptionPane.showOptionDialog(parentComponent, new Object[]{prompt, screen},
								"Calibration", JOptionPane.OK_CANCEL_OPTION,
								JOptionPane.QUESTION_MESSAGE, null, null, null);
			
			return new IntPoint(captureX, captureY);
		}
	}
	
	public double[][] makeGain(FloatImage src)
	{	
		double[][] red = src.getChannelData(0);
		double[][] green = src.getChannelData(1);
		double[][] blue = src.getChannelData(2);
		
		// now we're just searching for the largest gain
		double[][] result = new double[imgWidth][imgHeight];
		
		double gain = 0.0;
		
		// this is because the blurring does not blur the very edges
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				gain = red[i][j] + green[i][j] + blue[i][j] - baseline[i][j];
				
				result[i][j] = gain;
			}
		}
		
		return result;
	}
	
	public void makeAllBlack()
	{
		double[][] red = internalBuffer.getChannelData(0);
		double[][] green = internalBuffer.getChannelData(1);
		double[][] blue = internalBuffer.getChannelData(2);
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				red[i][j] = -1.0;
				green[i][j] = -1.0;
				blue[i][j] = -1.0;
			}
		}
	}
	
	public void calculateBaseline(FloatImage src)
	{
		baseline = new double[imgWidth][imgHeight];
		
		double[][] red = src.getChannelData(0);
		double[][] green = src.getChannelData(1);
		double[][] blue = src.getChannelData(2);
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				baseline[i][j] = red[i][j] + green[i][j] + blue[i][j];
			}
		}
	}
	
	public void buildTable()
	{
		transformTable = new IntPoint[imgWidth][imgHeight];
		
		int pdx, pdy, cdx, cdy;
		
		pdx = firstDotProj.x - secondDotProj.x;
		pdy = firstDotProj.y - secondDotProj.y;
		cdx = firstDotCam.x - secondDotCam.x;
		cdy = firstDotCam.y - secondDotCam.y;
		
		double pDistance = Math.sqrt((double)(pdx * pdx) + (double)(pdy * pdy));
		double cDistance = Math.sqrt((double)(cdx * cdx) + (double)(cdy * cdy));
		
		double scaleFactor = cDistance / pDistance;
		
		double scaleX = (double)cdx / (double)pdx;
		double scaleY = (double)cdy / (double)pdy;
		
		int ti, tj;
		int dx, dy, tdx, tdy;
		
		for(int i = 0; i < imgWidth; ++i)
		{
			for(int j = 0; j < imgHeight; ++j)
			{
				dx = i - firstDotProj.x;
				dy = j - firstDotProj.y;
				
				if(identifyMode == IDENTIFY_ALG)
				{
					tdx = (int)(dx * scaleFactor);
					tdy = (int)(dy * scaleFactor);
				}
				else
				{
					tdx = (int)(dx * scaleX);
					tdy = (int)(dy * scaleY);
				}
				
				ti = firstDotCam.x + tdx;
				tj = firstDotCam.y + tdy;
				
				if(ti >= 0 && tj >= 0 && ti < imgWidth && tj < imgHeight)
				{
					transformTable[i][j] = new IntPoint(ti, tj);
				}
			}
		}
	}
	
	public IntPoint[][] getTransform()
	{
		buildTable();
		return transformTable;
	}
	
	public void mouseClicked(MouseEvent e)
	{
		// do stuff
		if(e != null)
		{
			captureX = e.getX();
			captureY = e.getY();
			
			System.out.println("Mouse clicked at ( " + captureX + ", " + captureY + " )");
		}
	}
	
	public void mousePressed(MouseEvent e)
	{
		// don't do anything
	}
	
	public void mouseReleased(MouseEvent e)
	{
	}
	
	public void mouseEntered(MouseEvent e)
	{
	}
	
	public void mouseExited(MouseEvent e)
	{
	}
}